"""
Enhancing Piccolo integration with FastAPI.
"""

from __future__ import annotations

import typing as t
from collections import defaultdict
from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum
from inspect import Parameter, Signature

from fastapi import APIRouter, FastAPI, Request
from fastapi.params import Query
from pydantic import BaseModel as PydanticBaseModel
from pydantic.main import BaseModel

from piccolo_api.crud.endpoints import PiccoloCRUD

ANNOTATIONS: t.DefaultDict = defaultdict(dict)


class HTTPMethod(str, Enum):
    get = "GET"
    delete = "DELETE"


@dataclass
class FastAPIKwargs:
    """
    Allows kwargs to be passed into ``FastAPIApp.add_api_route``.
    """

    all_routes: t.Dict[str, t.Any] = field(default_factory=dict)
    get: t.Dict[str, t.Any] = field(default_factory=dict)
    delete: t.Dict[str, t.Any] = field(default_factory=dict)
    post: t.Dict[str, t.Any] = field(default_factory=dict)
    put: t.Dict[str, t.Any] = field(default_factory=dict)
    patch: t.Dict[str, t.Any] = field(default_factory=dict)
    get_single: t.Dict[str, t.Any] = field(default_factory=dict)
    delete_single: t.Dict[str, t.Any] = field(default_factory=dict)

    def get_kwargs(self, endpoint_name: str) -> t.Dict[str, t.Any]:
        """
        Merges the arguments for all routes with arguments specific to the
        given route.
        """
        default = self.all_routes.copy()
        route_specific = getattr(self, endpoint_name, {})
        default.update(**route_specific)
        return default


class CountModel(BaseModel):
    count: int
    page_size: int


class ReferenceModel(BaseModel):
    tableName: str
    columnName: str


class ReferencesModel(BaseModel):
    references: t.List[ReferenceModel]


class FastAPIWrapper:
    """
    Wraps ``PiccoloCRUD`` so it can easily be integrated into FastAPI.
    ``PiccoloCRUD`` can be used with any ASGI framework, but this way you get
    some of the benefits of FastAPI - namely, the OpenAPI integration. You get
    more control by building your own endpoints by hand, but ``FastAPIWrapper``
    works great for getting endpoints up and running very quickly, and reducing
    boilerplate code.

    :param root_url:
        The URL to mount the endpoint at - e.g. /movies/.
    :param fastapi_app:
        The ``FastAPI`` instance you want to attach the endpoints to.
    :param piccolo_crud:
        The ``PiccoloCRUD`` instance to wrap. ``FastAPIWrapper`` will obey
        the arguments passed into ``PiccoloCRUD``, for example ``ready_only``
        and ``allow_bulk_delete``.
    :param fastapi_kwargs:
        Specifies the extra kwargs to pass to FastAPI's `add_api_route`.

    """

    def __init__(
        self,
        root_url: str,
        fastapi_app: t.Union[FastAPI, APIRouter],
        piccolo_crud: PiccoloCRUD,
        fastapi_kwargs: FastAPIKwargs = FastAPIKwargs(),
    ):
        self.root_url = root_url
        self.fastapi_app = fastapi_app
        self.piccolo_crud = piccolo_crud
        self.fastapi_kwargs = fastapi_kwargs

        self.ModelOut = piccolo_crud.pydantic_model_output
        self.ModelIn = piccolo_crud.pydantic_model
        self.ModelOptional = piccolo_crud.pydantic_model_optional
        self.ModelPlural = piccolo_crud.pydantic_model_plural()

        self.alias = f"{piccolo_crud.table._meta.tablename}__{id(self)}"

        global ANNOTATIONS
        ANNOTATIONS[self.alias]["ModelIn"] = self.ModelIn
        ANNOTATIONS[self.alias]["ModelOut"] = self.ModelOut
        ANNOTATIONS[self.alias]["ModelOptional"] = self.ModelOptional
        ANNOTATIONS[self.alias]["ModelPlural"] = self.ModelPlural

        #######################################################################
        # Root - GET

        async def get(request: Request, **kwargs):
            """
            Returns all rows matching the given query.
            """
            return await piccolo_crud.root(request=request)

        self.modify_signature(
            endpoint=get,
            model=self.ModelOut,
            http_method=HTTPMethod.get,
            allow_ordering=True,
            allow_pagination=True,
        )

        fastapi_app.add_api_route(
            path=root_url,
            endpoint=get,
            methods=["GET"],
            response_model=self.ModelPlural,
            **fastapi_kwargs.get_kwargs("get"),
        )

        #######################################################################
        # Root - IDs

        async def ids(
            request: Request,
            search: t.Optional[str] = None,
            limit: t.Optional[int] = None,
        ):
            """
            Returns a mapping of row IDs to a readable representation.
            """
            return await piccolo_crud.get_ids(request=request)

        fastapi_app.add_api_route(
            path=self.join_urls(root_url, "/ids/"),
            endpoint=ids,
            methods=["GET"],
            response_model=t.Dict[str, str],
            **fastapi_kwargs.get_kwargs("get"),
        )

        #######################################################################
        # Root - New

        async def new(request: Request):
            """
            Returns all of the default values for a new row,
            but doesn't save it.
            """
            return await piccolo_crud.get_new(request=request)

        fastapi_app.add_api_route(
            path=self.join_urls(root_url, "/new/"),
            endpoint=new,
            methods=["GET"],
            response_model=t.Dict[str, str],
            **fastapi_kwargs.get_kwargs("get"),
        )

        #######################################################################
        # Root - Count

        async def count(request: Request, **kwargs):
            """
            Returns the number of rows matching the given query.
            """
            return await piccolo_crud.get_count(request=request)

        self.modify_signature(
            endpoint=count, model=self.ModelOut, http_method=HTTPMethod.get
        )

        fastapi_app.add_api_route(
            path=self.join_urls(root_url, "/count/"),
            endpoint=count,
            methods=["GET"],
            response_model=CountModel,
            **fastapi_kwargs.get_kwargs("get"),
        )

        #######################################################################
        # Root - Schema

        async def schema(request: Request):
            """
            Returns the JSON schema for the given table.
            """
            return await piccolo_crud.get_schema(request=request)

        fastapi_app.add_api_route(
            path=self.join_urls(root_url, "/schema/"),
            endpoint=schema,
            methods=["GET"],
            response_model=t.Dict[str, t.Any],
            **fastapi_kwargs.get_kwargs("get"),
        )

        #######################################################################
        # Root - References

        async def references(request: Request):
            """
            Returns a list of objects showing relationships with other tables.
            """
            return await piccolo_crud.get_references(request=request)

        fastapi_app.add_api_route(
            path=self.join_urls(root_url, "/references/"),
            endpoint=references,
            methods=["GET"],
            response_model=ReferencesModel,
            **fastapi_kwargs.get_kwargs("get"),
        )

        #######################################################################
        # Root - DELETE

        if not piccolo_crud.read_only and piccolo_crud.allow_bulk_delete:

            async def delete(request: Request, **kwargs):
                """
                Deletes all rows matching the given query.
                """
                return await piccolo_crud.root(request=request)

            self.modify_signature(
                endpoint=delete,
                model=self.ModelOut,
                http_method=HTTPMethod.delete,
            )

            fastapi_app.add_api_route(
                path=root_url,
                endpoint=delete,
                response_model=None,
                methods=["DELETE"],
                **fastapi_kwargs.get_kwargs("delete"),
            )

        #######################################################################
        # Root - POST

        if not piccolo_crud.read_only:

            async def post(request: Request, model):
                """
                Create a new row in the table.
                """
                return await piccolo_crud.root(request=request)

            post.__annotations__[
                "model"
            ] = f"ANNOTATIONS['{self.alias}']['ModelIn']"

            fastapi_app.add_api_route(
                path=root_url,
                endpoint=post,
                response_model=self.ModelOut,
                methods=["POST"],
                **fastapi_kwargs.get_kwargs("post"),
            )

        #######################################################################
        # Detail - GET

        async def get_single(row_id: int, request: Request):
            """
            Retrieve a single row from the table.
            """
            return await piccolo_crud.detail(request=request)

        fastapi_app.add_api_route(
            path=self.join_urls(root_url, "/{row_id:int}/"),
            endpoint=get_single,
            response_model=self.ModelOut,
            methods=["GET"],
            **fastapi_kwargs.get_kwargs("get_single"),
        )

        #######################################################################
        # Detail - DELETE

        if not piccolo_crud.read_only:

            async def delete_single(row_id: int, request: Request):
                """
                Delete a single row from the table.
                """
                return await piccolo_crud.detail(request=request)

            fastapi_app.add_api_route(
                path=self.join_urls(root_url, "/{row_id:int}/"),
                endpoint=delete_single,
                response_model=None,
                methods=["DELETE"],
                **fastapi_kwargs.get_kwargs("delete_single"),
            )

        #######################################################################
        # Detail - PUT

        if not piccolo_crud.read_only:

            async def put(row_id: int, request: Request, model):
                """
                Insert or update a single row.
                """
                return await piccolo_crud.detail(request=request)

            put.__annotations__[
                "model"
            ] = f"ANNOTATIONS['{self.alias}']['ModelIn']"

            fastapi_app.add_api_route(
                path=self.join_urls(root_url, "/{row_id:int}/"),
                endpoint=put,
                response_model=self.ModelOut,
                methods=["PUT"],
                **fastapi_kwargs.get_kwargs("put"),
            )

        #######################################################################
        # Detail - PATCH

        if not piccolo_crud.read_only:

            async def patch(row_id: int, request: Request, model):
                """
                Update a single row.
                """
                return await piccolo_crud.detail(request=request)

            patch.__annotations__[
                "model"
            ] = f"ANNOTATIONS['{self.alias}']['ModelOptional']"

            fastapi_app.add_api_route(
                path=self.join_urls(root_url, "/{row_id:int}/"),
                endpoint=patch,
                response_model=self.ModelOut,
                methods=["PATCH"],
                **fastapi_kwargs.get_kwargs("patch"),
            )

    @staticmethod
    def join_urls(url_1: str, url_2: str) -> str:
        """
        Combine two urls, and prevent double slashes (e.g. '/foo//bar')

        :param url_1:
            e.g. '/foo'
        :param url_2:
            e.g. '/bar
        :returns:
            e.g. '/foo/bar'

        """
        return "/".join([url_1.rstrip("/"), url_2.lstrip("/")])

    @staticmethod
    def modify_signature(
        endpoint: t.Callable,
        model: t.Type[PydanticBaseModel],
        http_method: HTTPMethod,
        allow_pagination: bool = False,
        allow_ordering: bool = False,
    ):
        """
        Modify the endpoint's signature, so FastAPI can correctly extract the
        schema from it. GET endpoints are given more filters.
        """
        parameters = [
            Parameter(
                name="request",
                kind=Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Request,
            ),
        ]

        for field_name, _field in model.__fields__.items():
            type_ = _field.outer_type_
            parameters.append(
                Parameter(
                    name=field_name,
                    kind=Parameter.POSITIONAL_OR_KEYWORD,
                    default=Query(
                        default=None,
                        description=(f"Filter by the `{field_name}` column."),
                    ),
                    annotation=type_,
                ),
            )

            if type_ in (int, float, Decimal):
                parameters.append(
                    Parameter(
                        name=f"{field_name}__operator",
                        kind=Parameter.POSITIONAL_OR_KEYWORD,
                        default=Query(
                            default=None,
                            description=(
                                f"Which operator to use for `{field_name}`. "
                                "The options are `e` (equals - default) `lt`, "
                                "`lte`, `gt`, and `gte`."
                            ),
                        ),
                    )
                )

            if type_ is str:
                parameters.append(
                    Parameter(
                        name=f"{field_name}__match",
                        kind=Parameter.POSITIONAL_OR_KEYWORD,
                        default=Query(
                            default=None,
                            description=(
                                f"Specifies how `{field_name}` should be "
                                "matched - `contains` (default), `exact`, "
                                "`starts`, `ends`."
                            ),
                        ),
                    )
                )

        if http_method == HTTPMethod.get:
            if allow_ordering:
                parameters.extend(
                    [
                        Parameter(
                            name="__order",
                            kind=Parameter.POSITIONAL_OR_KEYWORD,
                            annotation=str,
                            default=Query(
                                default=None,
                                description=(
                                    "Specifies which field to sort the "
                                    "results by. For example `id` to sort by "
                                    "id, and `-id` for descending."
                                ),
                            ),
                        )
                    ]
                )

            if allow_pagination:
                parameters.extend(
                    [
                        Parameter(
                            name="__page_size",
                            kind=Parameter.POSITIONAL_OR_KEYWORD,
                            annotation=int,
                            default=Query(
                                default=None,
                                description=(
                                    "The number of results to return."
                                ),
                            ),
                        ),
                        Parameter(
                            name="__page",
                            kind=Parameter.POSITIONAL_OR_KEYWORD,
                            annotation=int,
                            default=Query(
                                default=None,
                                description=(
                                    "Which page of results to return (default "
                                    "1)."
                                ),
                            ),
                        ),
                        Parameter(
                            name="__visible_fields",
                            kind=Parameter.POSITIONAL_OR_KEYWORD,
                            annotation=str,
                            default=Query(
                                default=None,
                                description=(
                                    "The fields to return. It's a comma "
                                    "separated list - for example "
                                    "'name,address'. By default all fields "
                                    "are returned."
                                ),
                            ),
                        ),
                    ]
                )

        endpoint.__signature__ = Signature(  # type: ignore
            parameters=parameters
        )



from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod

from piccolo.apps.user.tables import BaseUser as BaseUserTable
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
    BaseUser,
    SimpleUser,
)
from starlette.requests import HTTPConnection

from piccolo_api.shared.auth import User
from piccolo_api.token_auth.tables import TokenAuth


class TokenAuthProvider(metaclass=ABCMeta):
    """
    Subclass to create your own token provider.
    """

    @abstractmethod
    async def get_user(self, token: str) -> BaseUser:
        pass


class SecretTokenAuthProvider(TokenAuthProvider):
    """
    Checks that the token belongs to a predefined list of tokens. This is
    useful for very simple authentication use cases - such as internal
    microservices, where the client is trusted.
    """

    def __init__(self, tokens: t.Sequence[str]):
        self.tokens = tokens

    async def get_user(self, token: str) -> SimpleUser:
        if token in self.tokens:
            user = SimpleUser(username="secret_token_user")
            return user

        raise AuthenticationError("Token not recognised")


class PiccoloTokenAuthProvider(TokenAuthProvider):
    """
    Use this when the token is stored in a Piccolo database table.
    """

    def __init__(
        self,
        auth_table: t.Type[BaseUserTable] = BaseUserTable,
        token_table: t.Type[TokenAuth] = TokenAuth,
    ):
        self.auth_table = auth_table
        self.token_table = token_table

    async def get_user(self, token: str) -> User:
        user_id = await self.token_table.get_user_id(token)

        if not user_id:
            raise AuthenticationError()

        user = (
            await self.auth_table.select(self.auth_table.username)
            .where(self.auth_table._meta.primary_key == user_id)
            .first()
            .run()
        )

        if not user:
            raise AuthenticationError()

        user = User(user=user)
        return user


DEFAULT_PROVIDER = PiccoloTokenAuthProvider()


class TokenAuthBackend(AuthenticationBackend):
    def __init__(
        self,
        token_auth_provider: TokenAuthProvider = DEFAULT_PROVIDER,
    ):
        super().__init__()
        self.token_auth_provider = token_auth_provider

    def extract_token(self, header: str) -> str:
        try:
            token = header.split("Bearer ")[1]
        except IndexError:
            raise AuthenticationError("The header is in the wrong format.")

        return token

    async def authenticate(
        self, conn: HTTPConnection
    ) -> t.Optional[t.Tuple[AuthCredentials, BaseUser]]:

        auth_header = conn.headers.get("Authorization", None)

        if not auth_header:
            raise AuthenticationError("The Authorization header is missing.")

        token = self.extract_token(auth_header)

        user = await self.token_auth_provider.get_user(token=token)

        return (AuthCredentials(scopes=["authenticated"]), user)



import typing as t

from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
    BaseUser,
)
from starlette.requests import HTTPConnection


class AuthenticationBackendJunction(AuthenticationBackend):
    """
    Use when you want several different authentication backends to protect
    the same endpoint - if any of them pass, then auth is successful.
    """

    def __init__(self, backends: t.Sequence[AuthenticationBackend]):
        self.backends = backends

    async def authenticate(
        self, conn: HTTPConnection
    ) -> t.Optional[t.Tuple[AuthCredentials, BaseUser]]:

        for backend in self.backends:
            try:
                response = await backend.authenticate(conn=conn)
            except AuthenticationError:
                pass
            else:
                return response

        raise AuthenticationError("Auth failed on all backends.")



from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod

from piccolo.apps.user.tables import BaseUser
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from .tables import TokenAuth


class TokenProvider(metaclass=ABCMeta):
    """
    Subclass this to provide your own custom token provider.
    """

    @abstractmethod
    async def get_token(self, username: str, password: str) -> t.Optional[str]:
        pass


class PiccoloTokenProvider(TokenProvider):
    """
    Retrieves a token from a Piccolo table.
    """

    async def get_token(self, username: str, password: str) -> t.Optional[str]:
        user = await BaseUser.login(username=username, password=password)

        if user:
            response = (
                await TokenAuth.select(TokenAuth.token)
                .first()
                .where(TokenAuth.user == user)
                .run()
            )
            return response["token"]

        return None


class TokenAuthLoginEndpoint(HTTPEndpoint):

    token_provider: TokenProvider = PiccoloTokenProvider()

    async def post(self, request: Request) -> Response:
        """
        Return a token if the credentials are correct.
        """
        json = await request.json()
        username = json.get("username")
        password = json.get("password")

        if username and password:
            token = await self.token_provider.get_token(
                username=username, password=password
            )
            if token:
                return JSONResponse({"token": str(token)})
            else:
                return Response(
                    content="The credentials were incorrect",
                    status_code=401,
                )
        else:
            return Response(
                content="No credentials were found.", status_code=401
            )



import os
import typing as t

import jinja2
from fastapi.openapi.docs import get_swagger_ui_oauth2_redirect_html
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import HTMLResponse
from starlette.routing import Router

from piccolo_api.csrf.middleware import (
    DEFAULT_COOKIE_NAME,
    DEFAULT_HEADER_NAME,
)

ENVIRONMENT = jinja2.Environment(
    loader=jinja2.FileSystemLoader(
        searchpath=os.path.join(os.path.dirname(__file__), "templates")
    )
)


def swagger_ui(
    schema_url: str = "/openapi.json",
    swagger_ui_title: str = "Piccolo Swagger UI",
    csrf_cookie_name: t.Optional[str] = DEFAULT_COOKIE_NAME,
    csrf_header_name: t.Optional[str] = DEFAULT_HEADER_NAME,
):
    """
    Even though ASGI frameworks such as FastAPI and BlackSheep have endpoints
    for viewing OpenAPI / Swagger docs, out of the box they don't work well
    with some Piccolo middleware (namely CSRF middleware, which requires
    Swagger UI to add the CSRF token to the header of each API call).

    By using this endpoint instead, it will work correctly with CSRF.

    **FastAPI example**

    .. code-block:: python

        from fastapi import FastAPI
        from piccolo_api.openapi.endpoints import swagger_ui

        # By setting these values to None, we disable the builtin endpoints.
        app = FastAPI(docs_url=None, redoc_url=None)

        app.mount('/docs', swagger_ui())

    :param schema_url:
        The URL to the OpenAPI schema.
    :param csrf_cookie_name:
        The name of the CSRF cookie.
    :param csrf_header_name:
        The HTTP header name which the CSRF cookie value will be added to.

    """

    # We return a router, because it's effectively a mini ASGI
    # app, which can be mounted in any ASGI app which supports mounting.
    router = Router()

    class DocsEndpoint(HTTPEndpoint):
        def get(self, request: Request):
            template = ENVIRONMENT.get_template("swagger_ui.html.jinja")
            html = template.render(
                schema_url=schema_url,
                swagger_ui_title=swagger_ui_title,
                csrf_cookie_name=csrf_cookie_name,
                csrf_header_name=csrf_header_name,
            )
            return HTMLResponse(content=html)

    class OAuthRedirectEndpoint(HTTPEndpoint):
        def get(self, request: Request):
            return get_swagger_ui_oauth2_redirect_html()

    router.add_route("/", endpoint=DocsEndpoint)
    router.add_route("/oauth2-redirect/", endpoint=OAuthRedirectEndpoint)

    return router



from __future__ import annotations

import typing as t
from abc import abstractproperty
from datetime import datetime, timedelta

import jwt
from piccolo.apps.user.tables import BaseUser
from starlette.endpoints import HTTPEndpoint
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse


class JWTLoginBase(HTTPEndpoint):
    @abstractproperty
    def _auth_table(self) -> t.Type[BaseUser]:
        raise NotImplementedError

    @abstractproperty
    def _secret(self) -> str:
        raise NotImplementedError

    @abstractproperty
    def _expiry(self) -> timedelta:
        raise NotImplementedError

    async def post(self, request: Request) -> JSONResponse:
        body = await request.json()
        username = body.get("username", None)
        password = body.get("password", None)

        user_id = await self._auth_table.login(
            username=username, password=password
        )

        if not user_id:
            raise HTTPException(status_code=401, detail="Login failed")

        expiry = datetime.now() + self._expiry

        payload = jwt.encode({"user_id": user_id, "exp": expiry}, self._secret)

        return JSONResponse({"token": payload})


def jwt_login(
    secret: str,
    auth_table: t.Type[BaseUser] = BaseUser,
    expiry: timedelta = timedelta(days=1),
) -> t.Type[JWTLoginBase]:
    class JWTLogin(JWTLoginBase):
        _auth_table = auth_table
        _secret = secret
        _expiry = expiry

    return JWTLogin



from datetime import datetime

from .tables import SessionsBase


async def clean():
    """
    Removes any sessions from the table which have expired.
    """
    print("Removing old sessions ...")
    now = datetime.now()
    await SessionsBase.delete().where(SessionsBase.expiry_date < now).run()
    print("Successfully removed old sessions")



from __future__ import annotations

import typing as t
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from time import time

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import Response
from starlette.types import ASGIApp

if t.TYPE_CHECKING:  # pragma: no cover
    from starlette.middleware.base import Request, RequestResponseEndpoint


class RateLimitError(Exception):
    """
    Raised when a client exceeds the request limit. Should be handled
    internally without bleeding out to the rest of the application.
    """

    pass


class RateLimitProvider(metaclass=ABCMeta):
    """
    An abstract base class which all rate limit providers should inherit from.
    """

    @abstractmethod
    def increment(self, identifier: str):
        pass


class InMemoryLimitProvider(RateLimitProvider):
    """
    A very simple rate limiting provider - works fine when running a single
    application instance.

    Time values are given in seconds, rather than a timedelta, for improved
    performance.
    """

    def __init__(
        self,
        timespan: int,
        limit: int = 1000,
        block_duration: t.Optional[int] = None,
    ):
        """
        :param timespan:
            The time in seconds between resetting the number of requests.
            Beware setting it too high, because memory usage will increase.
        :param limit:
            The number of requests in the timespan, before getting blocked.
        :param block_duration:
            If set, the number of seconds before a client is no longer blocked.
            Otherwise, they're only removed when the app is restarted.
        """
        # Maps a client identifier to the number of requests they have made.
        self.request_dict: defaultdict = defaultdict(int)

        self.timespan = timespan
        self.last_reset = time()
        self.limit = limit

        self.blocked: t.Dict[str, float] = {}
        self.block_duration = block_duration

    def _handle_blocked(self):
        raise RateLimitError()

    def is_already_blocked(self, identifier: str) -> bool:
        """
        Check whether the identifier is already blocked from previous
        requests. Remove the identifier if the block has expired.
        """
        blocked_at: t.Optional[float] = self.blocked.get(identifier, None)
        if blocked_at:
            duration = self.block_duration
            if (time() - blocked_at < duration) if duration else True:
                return True
            else:
                del self.blocked[identifier]
                return False
        else:
            return False

    def add_to_blocked(self, identifier: str):
        self.blocked[identifier] = time()

    def increment(self, identifier: str):
        """
        Increment the number of requests with this identifier. If too many
        requests are received during the interval then record them as blocked,
        and reject the request.

        :param identifier:
            An identifier for the client making the request, for example the IP
            address.
        """
        if self.is_already_blocked(identifier):
            self._handle_blocked()

        # Reset the request count if needed.
        now = time()
        if now - self.last_reset > self.timespan:
            self.last_reset = now
            self.request_dict = defaultdict(int)

        self.request_dict[identifier] += 1

        if self.request_dict[identifier] > self.limit:
            self.add_to_blocked(identifier)
            self._handle_blocked()

    def clear_blocked(self):
        """
        Resets the block list.
        """
        self.blocked = {}


class RateLimitingMiddleware(BaseHTTPMiddleware):
    """
    Blocks clients who exceed a given number of requests in a given time
    period.
    """

    def __init__(
        self,
        app: ASGIApp,
        provider: RateLimitProvider = InMemoryLimitProvider(
            limit=1000, timespan=300
        ),
    ):
        super().__init__(app)
        self.rate_limit = provider

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        identifier = request.client.host
        try:
            self.rate_limit.increment(identifier)
        except RateLimitError:
            return Response(content="Too many requests", status_code=429)
        return await call_next(request)



import inspect
import typing as t
from enum import Enum

from piccolo.table import Table


class HookType(Enum):
    pre_save = "pre_save"
    pre_patch = "pre_patch"
    pre_delete = "pre_delete"


class Hook:
    def __init__(self, hook_type: HookType, callable: t.Callable) -> None:
        self.hook_type = hook_type
        self.callable = callable


async def execute_post_hooks(
    hooks: t.Dict[HookType, t.List[Hook]], hook_type: HookType, row: Table
):
    for hook in hooks.get(hook_type, []):
        if inspect.iscoroutinefunction(hook.callable):
            row = await hook.callable(row)
        else:
            row = hook.callable(row)
    return row


async def execute_patch_hooks(
    hooks: t.Dict[HookType, t.List[Hook]],
    hook_type: HookType,
    row_id: t.Any,
    values: t.Dict[t.Any, t.Any],
) -> t.Dict[t.Any, t.Any]:
    for hook in hooks.get(hook_type, []):
        if inspect.iscoroutinefunction(hook.callable):
            values = await hook.callable(row_id=row_id, values=values)
        else:
            values = hook.callable(row_id=row_id, values=values)
    return values


async def execute_delete_hooks(
    hooks: t.Dict[HookType, t.List[Hook]], hook_type: HookType, row_id: t.Any
):
    for hook in hooks.get(hook_type, []):
        if inspect.iscoroutinefunction(hook.callable):
            await hook.callable(row_id=row_id)
        else:
            hook.callable(row_id=row_id)



from __future__ import annotations

import typing as t
import uuid
from collections.abc import Sequence

from starlette.datastructures import URL
from starlette.middleware.base import (
    BaseHTTPMiddleware,
    Request,
    RequestResponseEndpoint,
)
from starlette.responses import Response
from starlette.types import ASGIApp

SAFE_HTTP_METHODS = ("GET", "HEAD", "OPTIONS", "TRACE")
ONE_YEAR = 31536000  # 365 * 24 * 60 * 60
DEFAULT_COOKIE_NAME = "csrftoken"
DEFAULT_HEADER_NAME = "X-CSRFToken"


class CSRFMiddleware(BaseHTTPMiddleware):
    """
    For GET requests, set a random token as a cookie. For unsafe HTTP methods,
    require a HTTP header to match the cookie value, otherwise the request
    is rejected.

    This uses the Double Submit Cookie style of CSRF prevention. For more
    information:

    https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
    https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#use-of-custom-request-headers

    By default, the CSRF token needs to be added to the request header. By
    setting ``allow_form_param`` to ``True``, it will also work if added as a
    form parameter.

    """

    @staticmethod
    def get_new_token() -> str:
        return str(uuid.uuid4())

    def __init__(
        self,
        app: ASGIApp,
        allowed_hosts: t.Sequence[str] = [],
        cookie_name: str = DEFAULT_COOKIE_NAME,
        header_name: str = DEFAULT_HEADER_NAME,
        max_age: int = ONE_YEAR,
        allow_header_param: bool = True,
        allow_form_param: bool = False,
        **kwargs,
    ):
        """
        :param app:
            The ASGI app you want to wrap.
        :allowed_hosts:
            If using this middleware with HTTPS, you need to set this value,
            for example ``['example.com']``.
        :cookie_name:
            You can specify a custom name for the cookie. There should be no
            need to change it, unless in the rare situation where the name
            clashes with another cookie.
        :header_name:
            You can tell the middleware to look for the CSRF token in a
            different HTTP header.
        :max_age:
            The max age of the cookie, in seconds.
        :allow_header_param:
            Whether to look for the CSRF token in the HTTP headers.
        :allow_form_param:
            Whether to look for the CSRF token in a form field with the same
            name as the cookie. By default, it's not enabled.

        """
        if not isinstance(allowed_hosts, Sequence):
            raise ValueError(
                "allowed_hosts must be a sequence (list or tuple)"
            )

        self.allowed_hosts = allowed_hosts
        self.cookie_name = cookie_name
        self.header_name = header_name
        self.max_age = max_age
        self.allow_header_param = allow_header_param
        self.allow_form_param = allow_form_param
        super().__init__(app, **kwargs)

    def is_valid_referer(self, request: Request) -> bool:
        header: str = (
            request.headers.get("origin")
            or request.headers.get("referer")
            or ""
        )

        url = URL(header)
        hostname = url.hostname
        is_valid = hostname in self.allowed_hosts if hostname else False
        return is_valid

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ):
        if request.method in SAFE_HTTP_METHODS:
            token = request.cookies.get(self.cookie_name, None)
            token_required = token is None

            if token_required:
                token = self.get_new_token()

            request.scope.update(
                {
                    "csrftoken": token,
                    "csrf_cookie_name": self.cookie_name,
                }
            )
            response = await call_next(request)

            if token_required and token:
                response.set_cookie(
                    self.cookie_name,
                    token,
                    max_age=self.max_age,
                )
            return response
        else:
            cookie_token = request.cookies.get(self.cookie_name)
            if not cookie_token:
                return Response("No CSRF cookie found", status_code=403)

            header_token = (
                request.headers.get(self.header_name)
                if self.allow_header_param
                else None
            )

            if self.allow_form_param and not header_token:
                form_data = await request.form()
                form_token = form_data.get(self.cookie_name, None)
                request.scope.update({"form": form_data})
            else:
                form_token = None

            if not header_token and not form_token:
                return Response(
                    "The CSRF token wasn't found in the form data or header.",
                    status_code=403,
                )

            if header_token and (cookie_token != header_token):
                return Response(
                    "The CSRF token in the header doesn't match the cookie.",
                    status_code=403,
                )

            if form_token and (cookie_token != form_token):
                return Response(
                    "The CSRF token in the form doesn't match the cookie.",
                    status_code=403,
                )

            # Provides defence in depth:
            if request.base_url.is_secure:
                # According to this paper, the referer header is present in
                # the vast majority of HTTPS requests, but not HTTP requests,
                # so only check it for HTTPS.
                # https://seclab.stanford.edu/websec/csrf/csrf.pdf
                if not self.is_valid_referer(request):
                    return Response(
                        "Referer or origin is incorrect", status_code=403
                    )

            request.scope.update(
                {
                    "csrftoken": cookie_token,
                    "csrf_cookie_name": self.cookie_name,
                }
            )

            return await call_next(request)



from __future__ import annotations

import functools
import inspect
import typing as t
from dataclasses import dataclass, field

from starlette.exceptions import HTTPException
from starlette.requests import Request

if t.TYPE_CHECKING:  # pragma: no cover
    from .endpoints import PiccoloCRUD


ValidatorFunction = t.Callable[["PiccoloCRUD", Request], None]


@dataclass
class Validators:
    """
    These validators are run by the corresponding method on PiccoloCRUD. Raise
    a Starlette exception if there is a problem.
    """

    every: t.List[ValidatorFunction] = field(default_factory=list)
    get_single: t.List[ValidatorFunction] = field(default_factory=list)
    put_single: t.List[ValidatorFunction] = field(default_factory=list)
    patch_single: t.List[ValidatorFunction] = field(default_factory=list)
    delete_single: t.List[ValidatorFunction] = field(default_factory=list)
    post_single: t.List[ValidatorFunction] = field(default_factory=list)
    get_all: t.List[ValidatorFunction] = field(default_factory=list)
    delete_all: t.List[ValidatorFunction] = field(default_factory=list)
    get_references: t.List[ValidatorFunction] = field(default_factory=list)
    get_ids: t.List[ValidatorFunction] = field(default_factory=list)
    get_new: t.List[ValidatorFunction] = field(default_factory=list)
    get_schema: t.List[ValidatorFunction] = field(default_factory=list)
    get_count: t.List[ValidatorFunction] = field(default_factory=list)
    extra_context: t.Dict[str, t.Any] = field(default_factory=dict)


def apply_validators(function):
    """
    A decorator used to apply validators to the corresponding methods on
    PiccoloCRUD.
    """

    def run_validators(*args, **kwargs):
        piccolo_crud: PiccoloCRUD = args[0]
        validators = piccolo_crud.validators

        request = kwargs.get("request") or next(
            (i for i in args if isinstance(i, Request)), None
        )

        validator_functions = (
            getattr(validators, function.__name__) + validators.every
        )
        if validator_functions and request:
            for validator_function in validator_functions:
                try:
                    validator_function(
                        request=request,
                        piccolo_crud=piccolo_crud,
                        **validators.extra_context,
                    )
                except HTTPException as exception:
                    raise exception
                except Exception:
                    raise HTTPException(
                        status_code=400, detail="Validation error"
                    )

    @functools.wraps(function)
    async def inner_coroutine_function(*args, **kwargs):
        run_validators(*args, **kwargs)
        return await function(*args, **kwargs)

    @functools.wraps(function)
    def inner_function(*args, **kwargs):
        run_validators(*args, **kwargs)
        return function(*args, **kwargs)

    if inspect.iscoroutinefunction(function):
        return inner_coroutine_function
    else:
        return inner_function



from __future__ import annotations

import itertools
import logging
import typing as t
from collections import defaultdict
from dataclasses import dataclass, field

import pydantic
from piccolo.columns import Column, Where
from piccolo.columns.column_types import Array, ForeignKey, Text, Varchar
from piccolo.columns.operators import (
    Equal,
    GreaterEqualThan,
    GreaterThan,
    LessEqualThan,
    LessThan,
)
from piccolo.columns.operators.comparison import ComparisonOperator
from piccolo.query.methods.delete import Delete
from piccolo.query.methods.select import Select
from piccolo.table import Table
from piccolo.utils.encoding import dump_json
from pydantic.error_wrappers import ValidationError
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.routing import Route, Router

from piccolo_api.crud.hooks import (
    Hook,
    HookType,
    execute_delete_hooks,
    execute_patch_hooks,
    execute_post_hooks,
)

from .exceptions import MalformedQuery
from .serializers import Config, create_pydantic_model
from .validators import Validators, apply_validators

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.query.methods.count import Count
    from piccolo.query.methods.objects import Objects
    from starlette.datastructures import QueryParams
    from starlette.routing import BaseRoute


logger = logging.getLogger(__file__)


OPERATOR_MAP = {
    "lt": LessThan,
    "lte": LessEqualThan,
    "gt": GreaterThan,
    "gte": GreaterEqualThan,
    "e": Equal,
}


MATCH_TYPES = ("contains", "exact", "starts", "ends")


class CustomJSONResponse(Response):
    media_type = "application/json"


@dataclass
class OrderBy:
    ascending: bool = False
    property_name: str = "id"


@dataclass
class Params:
    operators: t.Dict[str, t.Type[ComparisonOperator]] = field(
        default_factory=lambda: defaultdict(lambda: Equal)
    )
    match_types: t.Dict[str, str] = field(
        default_factory=lambda: defaultdict(lambda: MATCH_TYPES[0])
    )
    fields: t.Dict[str, t.Any] = field(default_factory=dict)
    order_by: t.Optional[OrderBy] = None
    include_readable: bool = False
    page: int = 1
    page_size: t.Optional[int] = None
    visible_fields: str = field(default="")


def get_visible_fields_options(
    table: t.Type[Table],
    exclude_secrets: bool = False,
    max_joins: int = 0,
    prefix: str = "",
) -> t.Tuple[str, ...]:
    """
    In the schema, we tell the user which fields are allowed with the
    ``__visible_fields`` GET parameter. This function extracts the column
    names, and names of related columns too.

    :param prefix:
        Used internally by this function - the user doesn't need to set this.

    """
    fields = []

    for column in table._meta.columns:
        if exclude_secrets and column._meta.secret:
            continue

        column_name = (
            f"{prefix}.{column._meta.name}" if prefix else column._meta.name
        )
        fields.append(column_name)

        if isinstance(column, ForeignKey) and max_joins > 0:
            fields.extend(
                get_visible_fields_options(
                    table=column._foreign_key_meta.resolved_references,
                    exclude_secrets=exclude_secrets,
                    max_joins=max_joins - 1,
                    prefix=column_name,
                )
            )

    return tuple(fields)


class PiccoloCRUD(Router):
    """
    Wraps a Piccolo table with CRUD methods for use in a REST API.
    """

    max_page_size: int = 1000

    def __init__(
        self,
        table: t.Type[Table],
        read_only: bool = True,
        allow_bulk_delete: bool = False,
        page_size: int = 15,
        exclude_secrets: bool = True,
        validators: Validators = Validators(),
        schema_extra: t.Optional[t.Dict[str, t.Any]] = None,
        max_joins: int = 0,
        hooks: t.Optional[t.List[Hook]] = None,
    ) -> None:
        """
        :param table:
            The Piccolo ``Table`` to expose CRUD methods for.
        :param read_only:
            If True, only the GET method is allowed.
        :param allow_bulk_delete:
            If True, allows a delete request to the root to delete all matching
            records. It is dangerous, so is disabled by default.
        :param page_size:
            The number of results shown on each page by default.
        :param exclude_secrets:
            Any Piccolo columns with ``secret=True`` will be omitted from the
            response.
        :param validators:
            Used to provide extra validation on certain endpoints - can be
            easier than subclassing.
        :param schema_extra:
            Additional information included in the Pydantic schema.
        :param max_joins:
            Determines whether a query can request data from related tables
            using joins. For example ``/movie/?__visible_fields=name,director.name``,
            which would return:

            .. code-block:: javascript

                {
                    'rows': [
                        {
                            'name': 'Star Wars',
                            'director': {
                                'name': 'George Lucas'
                            }
                        }
                    ]
                }

            This is a very powerful feature, but before enabling it, bear in
            mind the following:

             * If set too high, it could be used maliciously to craft slow
               queries which contain lots of joins, which could slow down your
               site.
             * Don't enable it if sensitive data is contained in related
               tables, as this feature can be used to retrieve that data.

            It's best used when the data in related tables is not of a
            sensitive nature and the client is highly trusted. Consider using
            it with ``exclude_secrets=True``.

            To see which fields can be filtered in this way, you can check
            the ``visible_fields_options`` value returned by the ``/schema``
            endpoint.

        """  # noqa: E501
        self.table = table
        self.page_size = page_size
        self.read_only = read_only
        self.allow_bulk_delete = allow_bulk_delete
        self.exclude_secrets = exclude_secrets
        self.validators = validators
        self.max_joins = max_joins
        if hooks:
            self._hook_map = {
                group[0]: [hook for hook in group[1]]
                for group in itertools.groupby(hooks, lambda x: x.hook_type)
            }
        else:
            self._hook_map = None  # type: ignore

        schema_extra = schema_extra if isinstance(schema_extra, dict) else {}
        self.visible_fields_options = get_visible_fields_options(
            table=table, exclude_secrets=exclude_secrets, max_joins=max_joins
        )
        schema_extra["visible_fields_options"] = self.visible_fields_options
        self.schema_extra = schema_extra

        root_methods = ["GET"]
        if not read_only:
            root_methods += (
                ["POST", "DELETE"] if allow_bulk_delete else ["POST"]
            )

        routes: t.List[BaseRoute] = [
            Route(path="/", endpoint=self.root, methods=root_methods),
            Route(
                path="/{row_id:int}/",
                endpoint=self.detail,
                methods=["GET"]
                if read_only
                else ["GET", "PUT", "DELETE", "PATCH"],
            ),
            Route(path="/schema/", endpoint=self.get_schema, methods=["GET"]),
            Route(path="/ids/", endpoint=self.get_ids, methods=["GET"]),
            Route(path="/count/", endpoint=self.get_count, methods=["GET"]),
            Route(
                path="/references/",
                endpoint=self.get_references,
                methods=["GET"],
            ),
            Route(path="/new/", endpoint=self.get_new, methods=["GET"]),
            Route(
                path="/password/",
                endpoint=self.update_password,
                methods=["PUT"],
            ),
        ]

        super().__init__(routes=routes)

    ###########################################################################

    @property
    def pydantic_model(self) -> t.Type[pydantic.BaseModel]:
        """
        Useful for serialising inbound data from POST and PUT requests.
        """
        return create_pydantic_model(
            self.table,
            model_name=f"{self.table.__name__}In",
            **self.schema_extra,
        )

    def _pydantic_model_output(
        self,
        include_readable: bool = False,
        include_columns: t.Tuple[Column, ...] = (),
        nested: t.Union[bool, t.Tuple[Column, ...]] = False,
    ) -> t.Type[pydantic.BaseModel]:
        return create_pydantic_model(
            self.table,
            include_default_columns=True,
            include_readable=include_readable,
            include_columns=include_columns,
            model_name=f"{self.table.__name__}Output",
            nested=nested,
        )

    @property
    def pydantic_model_output(self) -> t.Type[pydantic.BaseModel]:
        """
        Contains the default columns, which is required when exporting
        data (for example, in a GET request).
        """
        return self._pydantic_model_output()

    @property
    def pydantic_model_optional(self) -> t.Type[pydantic.BaseModel]:
        """
        All fields are optional, which is useful for serialising filters,
        where a user can filter on any number of fields.
        """
        return create_pydantic_model(
            self.table,
            include_default_columns=True,
            all_optional=True,
            model_name=f"{self.table.__name__}Optional",
        )

    def pydantic_model_plural(
        self,
        include_readable=False,
        include_columns: t.Tuple[Column, ...] = (),
        nested: t.Union[bool, t.Tuple[Column, ...]] = False,
    ):
        """
        This is for when we want to serialise many copies of the model.
        """
        base_model: t.Any = create_pydantic_model(
            self.table,
            include_default_columns=True,
            include_readable=include_readable,
            include_columns=include_columns,
            model_name=f"{self.table.__name__}Item",
            nested=nested,
        )
        return pydantic.create_model(
            str(self.table.__name__) + "Plural",
            __config__=Config,
            rows=(t.List[base_model], None),
        )

    @apply_validators
    async def get_schema(self, request: Request) -> JSONResponse:
        """
        Return a representation of the model, so a UI can generate a form.
        """
        return JSONResponse(self.pydantic_model.schema())

    ###########################################################################

    async def update_password(self, request: Request) -> Response:
        """
        Used to update password fields.
        """
        return Response("Coming soon", status_code=501)

    ###########################################################################

    @apply_validators
    async def get_ids(self, request: Request) -> Response:
        """
        Returns all the IDs for the current table, mapped to a readable
        representation e.g. {'1': 'joebloggs'}. Used for UI, like foreign
        key selectors.

        An optional 'search' GET parameter can be used to filter the results
        returned. Also, an optional 'limit' paramter can be used to specify
        how many results should be returned, and 'offset' for basic pagination.

        """
        readable = self.table.get_readable()
        query = self.table.select().columns(
            self.table._meta.primary_key._meta.name, readable
        )

        limit = request.query_params.get("limit")
        if limit is not None:
            try:
                limit = int(limit)
            except ValueError:
                return Response(
                    "The limit must be an integer", status_code=400
                )
        else:
            limit = "ALL"

        offset = request.query_params.get("offset")
        if offset is not None:
            try:
                offset = int(offset)
            except ValueError:
                return Response(
                    "The offset must be an integer", status_code=400
                )
        else:
            offset = 0

        search_term = request.query_params.get("search")
        if search_term is not None:
            # Readable doesn't currently have a 'like' method, so we do it
            # manually.
            if self.table._meta.db.engine_type == "postgres":
                query = t.cast(
                    Select,
                    self.table.raw(
                        (
                            f"SELECT * FROM ({query.__str__()}) as subquery "
                            "WHERE subquery.readable ILIKE {} "
                            f"LIMIT {limit} OFFSET {offset}"
                        ),
                        f"%{search_term}%",
                    ),
                )
            if self.table._meta.db.engine_type == "sqlite":
                # The conversion to uppercase is necessary as SQLite doesn't
                # support ILIKE.
                sql = (
                    f"SELECT * FROM ({query.__str__()}) as subquery "
                    "WHERE UPPER(subquery.readable) LIKE {}"
                )
                if isinstance(limit, int):
                    sql += f" LIMIT {limit} OFFSET {offset}"
                query = t.cast(
                    Select, self.table.raw(sql, f"%{search_term.upper()}%")
                )
        else:
            if limit != "ALL":
                query = query.limit(limit).offset(offset)

        values = await query.run()
        return JSONResponse({i["id"]: i["readable"] for i in values})

    ###########################################################################

    @apply_validators
    async def get_references(self, request: Request) -> JSONResponse:
        """
        Returns a list of tables with foreign keys to this table, along with
        the name of the foreign key column.
        """
        references = [
            {
                "tableName": i._meta.table._meta.tablename,
                "columnName": i._meta.name,
            }
            for i in self.table._meta.foreign_key_references
        ]
        return JSONResponse({"references": references})

    ###########################################################################

    @apply_validators
    async def get_count(self, request: Request) -> Response:
        """
        Returns the total number of rows in the table.
        """
        params = self._parse_params(request.query_params)
        split_params = self._split_params(params)

        try:
            query = self._apply_filters(self.table.count(), split_params)
        except MalformedQuery as exception:
            return Response(str(exception), status_code=400)

        count = await query.run()
        return JSONResponse({"count": count, "page_size": self.page_size})

    ###########################################################################

    def _parse_params(self, params: QueryParams) -> t.Dict[str, t.Any]:
        """
        The GET params may contain multiple values for each parameter name.
        For example:

        /tables/movie?tag=horror&tag=scifi

        Some clients, such as Axios, will use this convention:

        /tables/movie?tag[]=horror&tag[]=scifi

        This method normalises the parameter name, removing square brackets
        if present (tag[] -> tag), and will return a list of values if
        multiple are present.

        """
        params_map: t.Dict[str, t.Any] = {
            i[0]: [j[1] for j in i[1]]
            for i in itertools.groupby(params.multi_items(), lambda x: x[0])
        }

        array_columns = [
            i._meta.name
            for i in self.table._meta.columns
            if i.value_type == list
        ]

        output = {}

        for key, value in params_map.items():
            if key.endswith("[]") or key.rstrip("[]") in array_columns:
                # Is either an array, or multiple values have been passed in
                # for another field.
                key = key.rstrip("[]")
            elif len(value) == 1:
                value = value[0]

            output[key] = value

        return output

    async def root(self, request: Request) -> Response:
        if request.method == "GET":
            params = self._parse_params(request.query_params)
            return await self.get_all(request, params=params)
        elif request.method == "POST":
            data = await request.json()
            return await self.post_single(request, data)
        elif request.method == "DELETE":
            params = dict(request.query_params)
            return await self.delete_all(request, params=params)
        else:
            return Response(status_code=405)

    ###########################################################################

    @staticmethod
    def _split_params(params: t.Dict[str, t.Any]) -> Params:
        """
        Some parameters reference fields, and others provide instructions
        on how to perform the query (e.g. which operator to use).

        An example of an operator parameter is {'age__operator': 'gte'}.

        You can specify how to match text fields:
        {'name__match': 'exact'}.

        Ordering is specified like: {'__order': '-name'}.

        To include readable representations of foreign keys, use:
        {'__readable': 'true'}.

        For pagination, you can override the default page size:
        {'__page_size': 15}.

        And can specify which page: {'__page': 2}.

        You can specify which fields want to display in rows:
        {'__visible_fields': 'id,name'}.

        This method splits the params into their different types.
        """
        response = Params()

        for key, value in params.items():
            if key.endswith("__operator") and value in OPERATOR_MAP.keys():
                field_name = key.split("__operator")[0]
                response.operators[field_name] = OPERATOR_MAP[value]
                continue

            if key.endswith("__match") and value in MATCH_TYPES:
                field_name = key.split("__match")[0]
                response.match_types[field_name] = value
                continue

            if key == "__order":
                ascending = True
                if value.startswith("-"):
                    ascending = False
                    value = value[1:]
                response.order_by = OrderBy(
                    ascending=ascending, property_name=value
                )
                continue

            if key == "__page":
                try:
                    page = int(value)
                except ValueError:
                    logger.info(f"Unrecognised __page argument - {value}")
                else:
                    response.page = page
                continue

            if key == "__page_size":
                try:
                    page_size = int(value)
                except ValueError:
                    logger.info(f"Unrecognised __page_size argument - {value}")
                else:
                    response.page_size = page_size
                continue

            if key == "__visible_fields":
                response.visible_fields = value
                continue

            if key == "__readable" and value in ("true", "True", "1"):
                response.include_readable = True
                continue

            response.fields[key] = value

        return response

    def _apply_filters(
        self, query: t.Union[Select, Count, Objects, Delete], params: Params
    ) -> t.Union[Select, Count, Objects, Delete]:
        """
        Apply the HTTP query parameters to the Piccolo query object, then
        return it.

        Works on any queries which support `where` clauses - Select, Count,
        Objects etc.
        """
        fields = params.fields

        if fields:
            model_dict = self.pydantic_model_optional(**fields).dict()
            for field_name in fields.keys():
                value = model_dict.get(field_name, ...)
                if value is ...:
                    raise MalformedQuery(
                        f"{field_name} isn't a valid field name."
                    )
                column: Column = getattr(self.table, field_name)

                # Sometimes a list of values is passed in.
                values = value if isinstance(value, list) else [value]

                for value in values:
                    if isinstance(column, (Varchar, Text)):
                        match_type = params.match_types[field_name]
                        if match_type == "exact":
                            clause = column.__eq__(value)
                        elif match_type == "starts":
                            clause = column.ilike(f"{value}%")
                        elif match_type == "ends":
                            clause = column.ilike(f"%{value}")
                        else:
                            clause = column.ilike(f"%{value}%")
                        query = query.where(clause)
                    elif isinstance(column, Array):
                        query = query.where(column.any(value))
                    else:
                        operator = params.operators[field_name]
                        query = query.where(
                            Where(
                                column=column, value=value, operator=operator
                            )
                        )
        return query

    @apply_validators
    async def get_all(
        self, request: Request, params: t.Optional[t.Dict[str, t.Any]] = None
    ) -> Response:
        """
        Get all rows - query parameters are used for filtering.
        """
        params = self._clean_data(params) if params else {}

        split_params = self._split_params(params)

        # Visible fields
        visible_fields = split_params.visible_fields
        nested: t.Union[bool, t.Tuple[Column, ...]]
        if visible_fields:
            try:
                visible_columns = self._parse_visible_fields(visible_fields)
            except ValueError as exception:
                return Response(str(exception), status_code=400)

            nested = tuple(
                i for i in visible_columns if len(i._meta.call_chain) > 0
            )
        else:
            visible_columns = self.table._meta.columns
            nested = False

        # Readable
        include_readable = split_params.include_readable
        readable_columns = (
            [
                self.table._get_related_readable(i)
                for i in visible_columns
                if isinstance(i, ForeignKey)
            ]
            if include_readable
            else []
        )

        # Build select query, and exclude secrets
        query = self.table.select(
            *visible_columns,
            *readable_columns,
            exclude_secrets=self.exclude_secrets,
        )

        # Make it nested if required
        if nested:
            query = query.output(nested=True)

        # Apply filters
        try:
            query = t.cast(Select, self._apply_filters(query, split_params))
        except MalformedQuery as exception:
            return Response(str(exception), status_code=400)

        # Ordering
        order_by = split_params.order_by
        if order_by:
            column = getattr(self.table, order_by.property_name)
            query = query.order_by(column, ascending=order_by.ascending)
        else:
            query = query.order_by(
                self.table._meta.primary_key, ascending=False
            )

        # Pagination
        page_size = split_params.page_size or self.page_size
        # If the page_size is greater than max_page_size return an error
        if page_size > self.max_page_size:
            return JSONResponse(
                {"error": "The page size limit has been exceeded"},
                status_code=403,
            )
        query = query.limit(page_size)
        page = split_params.page
        if page > 1:
            offset = page_size * (page - 1)
            query = query.offset(offset).limit(page_size)

        rows = await query.run()

        # We need to serialise it ourselves, in case there are datetime
        # fields.
        json = self.pydantic_model_plural(
            include_readable=include_readable,
            include_columns=tuple(visible_columns),
            nested=nested,
        )(rows=rows).json()
        return CustomJSONResponse(json)

    ###########################################################################

    def _clean_data(self, data: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        cleaned_data: t.Dict[str, t.Any] = {}

        for key, value in data.items():
            value = (
                None
                if (isinstance(value, str) and value.lower() == "null")
                else value
            )
            cleaned_data[key] = value

        return cleaned_data

    @apply_validators
    async def post_single(
        self, request: Request, data: t.Dict[str, t.Any]
    ) -> Response:
        """
        Adds a single row, if the id doesn't already exist.
        """
        cleaned_data = self._clean_data(data)
        try:
            model = self.pydantic_model(**cleaned_data)
        except ValidationError as exception:
            return Response(str(exception), status_code=400)

        try:
            row = self.table(**model.dict())
            if self._hook_map:
                row = await execute_post_hooks(
                    hooks=self._hook_map, hook_type=HookType.pre_save, row=row
                )
            response = await row.save().run()
            json = dump_json(response)
            # Returns the id of the inserted row.
            return CustomJSONResponse(json, status_code=201)
        except ValueError:
            return Response("Unable to save the resource.", status_code=500)

    @apply_validators
    async def delete_all(
        self, request: Request, params: t.Optional[t.Dict[str, t.Any]] = None
    ) -> Response:
        """
        Deletes all rows - query parameters are used for filtering.
        """
        params = self._clean_data(params) if params else {}
        split_params = self._split_params(params)

        try:
            query = self._apply_filters(
                self.table.delete(force=True), split_params
            )
        except MalformedQuery as exception:
            return Response(str(exception), status_code=400)

        await query.run()
        return Response(status_code=204)

    ###########################################################################

    @apply_validators
    async def get_new(self, request: Request) -> CustomJSONResponse:
        """
        This endpoint is used when creating new rows in a UI. It provides
        all of the default values for a new row, but doesn't save it.
        """
        row = self.table(ignore_missing=True)
        row_dict = row.__dict__
        row_dict.pop("id", None)

        return CustomJSONResponse(
            self.pydantic_model_optional(**row_dict).json()
        )

    ###########################################################################

    async def detail(self, request: Request) -> Response:
        """
        If a resource with a matching ID isn't found, a 404 is returned.

        This is also the case for PUT requests - we don't want the user to be
        able to specify the ID of a new resource, as this could potentially
        cause issues.
        """
        row_id = request.path_params.get("row_id", None)
        if row_id is None:
            return Response("Missing ID parameter.", status_code=404)

        if (
            not await self.table.exists()
            .where(self.table._meta.primary_key == row_id)
            .run()
        ):
            return Response("The resource doesn't exist", status_code=404)

        if (type(row_id) is int) and row_id < 1:
            return Response(
                "The resource ID must be greater than 0", status_code=400
            )

        if request.method == "GET":
            return await self.get_single(request, row_id)
        elif request.method == "PUT":
            data = await request.json()
            return await self.put_single(request, row_id, data)
        elif request.method == "DELETE":
            return await self.delete_single(request, row_id)
        elif request.method == "PATCH":
            data = await request.json()
            return await self.patch_single(request, row_id, data)
        else:
            return Response(status_code=405)

    def _parse_visible_fields(self, visible_fields: str) -> t.List[Column]:
        """
        Parse the ``visible_fields`` string, and return a list of columns.

        :param visible_fields:
            A comma separated list of column names, for example ``'id,name'``.
            The presence of a full stop in the name indicates a join, for
            example ``'director.name'``.
        :raises ValueError:
            If the max join depth is exceeded, or the column name isn't
            recognised.

        """
        column_names: t.List[str] = visible_fields.split(",")
        visible_columns: t.List[Column] = []

        for column_name in column_names:
            try:
                column = self.table._meta.get_column_by_name(column_name)
            except ValueError as exception:
                raise ValueError(
                    f"{exception} - the column options are "
                    f"{self.visible_fields_options}."
                )

            if len(column._meta.call_chain) > self.max_joins:
                raise ValueError("Max join depth exceeded")
            else:
                visible_columns.append(column)

        return visible_columns

    @apply_validators
    async def get_single(self, request: Request, row_id: int) -> Response:
        """
        Returns a single row.
        """
        params = dict(request.query_params)
        split_params: Params = self._split_params(params)

        # Visible fields
        nested: t.Union[bool, t.Tuple[Column, ...]]
        visible_fields = split_params.visible_fields
        if visible_fields:
            try:
                visible_columns = self._parse_visible_fields(visible_fields)
            except ValueError as exception:
                return Response(str(exception), status_code=400)

            nested = tuple(
                i for i in visible_columns if len(i._meta.call_chain) > 0
            )
        else:
            visible_columns = self.table._meta.columns
            nested = False

        # Readable
        readable_columns = (
            [
                self.table._get_related_readable(i)
                for i in self.table._meta.foreign_key_columns
            ]
            if split_params.include_readable
            else []
        )

        query = (
            self.table.select(
                *visible_columns,
                *readable_columns,
                exclude_secrets=self.exclude_secrets,
            )
            .where(self.table._meta.primary_key == row_id)
            .first()
        )

        if nested:
            query = query.output(nested=True)

        row = await query.run()

        if not row:
            return Response(
                "Unable to find a resource with that ID.", status_code=404
            )

        return CustomJSONResponse(
            self._pydantic_model_output(
                include_readable=split_params.include_readable,
                include_columns=tuple(visible_columns),
                nested=nested,
            )(**row).json()
        )

    @apply_validators
    async def put_single(
        self, request: Request, row_id: int, data: t.Dict[str, t.Any]
    ) -> Response:
        """
        Replaces an existing row. We don't allow new resources to be created.
        """
        cleaned_data = self._clean_data(data)

        try:
            model = self.pydantic_model(**cleaned_data)
        except ValidationError as exception:
            return Response(str(exception), status_code=400)

        cls = self.table

        values = {
            getattr(cls, key): getattr(model, key) for key in data.keys()
        }

        try:

            await cls.update(values).where(
                cls._meta.primary_key == row_id
            ).run()
            return Response(status_code=204)
        except ValueError:
            return Response("Unable to save the resource.", status_code=500)

    @apply_validators
    async def patch_single(
        self, request: Request, row_id: int, data: t.Dict[str, t.Any]
    ) -> Response:
        """
        Patch a single row.
        """
        cleaned_data = self._clean_data(data)

        try:
            model = self.pydantic_model_optional(**cleaned_data)
        except ValidationError as exception:
            return Response(str(exception), status_code=400)

        cls = self.table

        try:
            values = {
                getattr(cls, key): getattr(model, key) for key in data.keys()
            }
        except AttributeError:
            unrecognised_keys = set(data.keys()) - set(model.dict().keys())
            return Response(
                f"Unrecognised keys - {unrecognised_keys}.", status_code=400
            )

        if self._hook_map:
            values = await execute_patch_hooks(
                hooks=self._hook_map,
                hook_type=HookType.pre_patch,
                row_id=row_id,
                values=values,
            )

        try:
            await cls.update(values).where(
                cls._meta.primary_key == row_id
            ).run()
            new_row = (
                await cls.select(exclude_secrets=self.exclude_secrets)
                .where(cls._meta.primary_key == row_id)
                .first()
                .run()
            )
            return JSONResponse(self.pydantic_model(**new_row).json())
        except ValueError:
            return Response("Unable to save the resource.", status_code=500)

    @apply_validators
    async def delete_single(self, request: Request, row_id: int) -> Response:
        """
        Deletes a single row.
        """

        if self._hook_map:
            await execute_delete_hooks(
                hooks=self._hook_map,
                hook_type=HookType.pre_delete,
                row_id=row_id,
            )

        try:
            await self.table.delete().where(
                self.table._meta.primary_key == row_id
            ).run()
            return Response(status_code=204)
        except ValueError:
            return Response("Unable to delete the resource.", status_code=500)


__all__ = ["PiccoloCRUD"]



class MalformedQuery(Exception):
    """
    Raised when the query is malformed - for example, the column names are
    unrecognised. The exception should be handled internally by PiccoloCRUD,
    and shouldn't bleed out to the wider application.
    """

    pass



from __future__ import annotations

import os
import typing as t
import warnings
from abc import ABCMeta, abstractproperty
from datetime import datetime, timedelta
from json import JSONDecodeError

from jinja2 import Environment, FileSystemLoader
from piccolo.apps.user.tables import BaseUser
from starlette.endpoints import HTTPEndpoint, Request
from starlette.exceptions import HTTPException
from starlette.responses import (
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
)
from starlette.status import HTTP_303_SEE_OTHER

from piccolo_api.session_auth.tables import SessionsBase

if t.TYPE_CHECKING:  # pragma: no cover
    from jinja2 import Template
    from starlette.responses import Response


LOGIN_TEMPLATE_PATH = os.path.join(
    os.path.dirname(__file__), "templates", "login.html"
)

LOGOUT_TEMPLATE_PATH = os.path.join(
    os.path.dirname(__file__), "templates", "logout.html"
)


class SessionLogoutEndpoint(HTTPEndpoint, metaclass=ABCMeta):
    @abstractproperty
    def _session_table(self) -> t.Type[SessionsBase]:
        raise NotImplementedError

    @abstractproperty
    def _cookie_name(self) -> str:
        raise NotImplementedError

    @abstractproperty
    def _redirect_to(self) -> t.Optional[str]:
        raise NotImplementedError

    @abstractproperty
    def _logout_template(self) -> Template:
        raise NotImplementedError

    def render_template(
        self, request: Request, template_context: t.Dict[str, t.Any] = {}
    ) -> HTMLResponse:
        # If CSRF middleware is present, we have to include a form field with
        # the CSRF token. It only works if CSRFMiddleware has
        # allow_form_param=True, otherwise it only looks for the token in the
        # header.
        csrftoken = request.scope.get("csrftoken")
        csrf_cookie_name = request.scope.get("csrf_cookie_name")

        return HTMLResponse(
            self._logout_template.render(
                csrftoken=csrftoken,
                csrf_cookie_name=csrf_cookie_name,
                request=request,
                **template_context,
            )
        )

    async def get(self, request: Request) -> HTMLResponse:
        return self.render_template(request)

    async def post(self, request: Request) -> Response:
        cookie = request.cookies.get(self._cookie_name, None)
        if not cookie:
            raise HTTPException(
                status_code=401, detail="The session cookie wasn't found."
            )
        await self._session_table.remove_session(token=cookie)

        if self._redirect_to is not None:
            response: Response = RedirectResponse(
                url=self._redirect_to, status_code=HTTP_303_SEE_OTHER
            )
        else:
            response = PlainTextResponse("Successfully logged out")

        response.set_cookie(self._cookie_name, "", max_age=0)
        return response


class SessionLoginEndpoint(HTTPEndpoint, metaclass=ABCMeta):
    @abstractproperty
    def _auth_table(self) -> t.Type[BaseUser]:
        raise NotImplementedError

    @abstractproperty
    def _session_table(self) -> t.Type[SessionsBase]:
        raise NotImplementedError

    @abstractproperty
    def _session_expiry(self) -> timedelta:
        raise NotImplementedError

    @abstractproperty
    def _max_session_expiry(self) -> timedelta:
        raise NotImplementedError

    @abstractproperty
    def _cookie_name(self) -> str:
        raise NotImplementedError

    @abstractproperty
    def _redirect_to(self) -> t.Optional[str]:
        """
        Where to redirect to after login is successful.
        """
        raise NotImplementedError

    @abstractproperty
    def _production(self) -> bool:
        """
        If True, apply more stringent security.
        """
        raise NotImplementedError

    @abstractproperty
    def _login_template(self) -> Template:
        raise NotImplementedError

    def render_template(
        self, request: Request, template_context: t.Dict[str, t.Any] = {}
    ) -> HTMLResponse:
        # If CSRF middleware is present, we have to include a form field with
        # the CSRF token. It only works if CSRFMiddleware has
        # allow_form_param=True, otherwise it only looks for the token in the
        # header.
        csrftoken = request.scope.get("csrftoken")
        csrf_cookie_name = request.scope.get("csrf_cookie_name")

        return HTMLResponse(
            self._login_template.render(
                csrftoken=csrftoken,
                csrf_cookie_name=csrf_cookie_name,
                request=request,
                **template_context,
            )
        )

    async def get(self, request: Request) -> HTMLResponse:
        return self.render_template(request)

    async def post(self, request: Request) -> Response:
        # Some middleware (for example CSRF) has already awaited the request
        # body, and adds it to the request.
        body = request.scope.get("form")

        if not body:
            try:
                body = await request.json()
            except JSONDecodeError:
                body = await request.form()

        username = body.get("username", None)
        password = body.get("password", None)

        if (not username) or (not password):
            raise HTTPException(
                status_code=401, detail="Missing username or password"
            )

        user_id = await self._auth_table.login(
            username=username, password=password
        )

        if not user_id:
            if body.get("format") == "html":
                return self.render_template(
                    request,
                    template_context={
                        "error": "The username or password is incorrect."
                    },
                )
            else:
                raise HTTPException(status_code=401, detail="Login failed")

        now = datetime.now()
        expiry_date = now + self._session_expiry
        max_expiry_date = now + self._max_session_expiry

        session: SessionsBase = await self._session_table.create_session(
            user_id=user_id,
            expiry_date=expiry_date,
            max_expiry_date=max_expiry_date,
        )

        if self._redirect_to is not None:
            response: Response = RedirectResponse(
                url=self._redirect_to, status_code=HTTP_303_SEE_OTHER
            )
        else:
            response = JSONResponse(
                content={"message": "logged in"}, status_code=200
            )

        if not self._production:
            message = (
                "If running sessions in production, make sure 'production' "
                "is set to True, and serve under HTTPS."
            )
            warnings.warn(message)

        cookie_value = t.cast(str, session.token)

        response.set_cookie(
            key=self._cookie_name,
            value=cookie_value,
            httponly=True,
            secure=self._production,
            max_age=int(self._max_session_expiry.total_seconds()),
            samesite="lax",
        )
        return response


def session_login(
    auth_table: t.Type[BaseUser] = BaseUser,
    session_table: t.Type[SessionsBase] = SessionsBase,
    session_expiry: timedelta = timedelta(hours=1),
    max_session_expiry: timedelta = timedelta(days=7),
    redirect_to: t.Optional[str] = "/",
    production: bool = False,
    cookie_name: str = "id",
    template_path: t.Optional[str] = None,
) -> t.Type[SessionLoginEndpoint]:
    """
    An endpoint for creating a user session.

    :param auth_table:
        Which table to authenticate the username and password with.
    :param session_table:
        Which table to store the session in.
    :param session_expiry:
        How long the session will last.
    :param max_session_expiry:
        If the session is refreshed (see the ``increase_expiry`` parameter for
        ``SessionsAuthBackend``), it can only be refreshed up to a certain
        limit, after which the session is void.
    :param redirect_to:
        Where to redirect to after successful login.
    :param production:
        Adds additional security measures. Use this in production, when serving
        your app over HTTPS.
    :param cookie_name:
        The name of the cookie used to store the session token. Only override
        this if the name of the cookie clashes with other cookies.
    :param template_path:
        If you want to override the default login HTML template, you can do
        so by specifying the absolute path to a custom template. For example
        ``'/some_directory/login.html'``. Refer to the default template at
        ``piccolo_api/session_auth/templates/login.html`` as a basis for your
        custom template.

    """
    template_path = (
        LOGIN_TEMPLATE_PATH if template_path is None else template_path
    )

    directory, filename = os.path.split(template_path)
    environment = Environment(loader=FileSystemLoader(directory))
    login_template = environment.get_template(filename)

    class _SessionLoginEndpoint(SessionLoginEndpoint):
        _auth_table = auth_table
        _session_table = session_table
        _session_expiry = session_expiry
        _max_session_expiry = max_session_expiry
        _redirect_to = redirect_to
        _production = production
        _cookie_name = cookie_name
        _login_template = login_template

    return _SessionLoginEndpoint


def session_logout(
    session_table: t.Type[SessionsBase] = SessionsBase,
    cookie_name: str = "id",
    redirect_to: t.Optional[str] = None,
    template_path: t.Optional[str] = None,
) -> t.Type[SessionLogoutEndpoint]:
    """
    An endpoint for clearing a user session.

    :param session_table:
        Which table to store the session in.
    :param cookie_name:
        The name of the cookie used to store the session token. Only override
        this if the name of the cookie clashes with other cookies.
    :param redirect_to:
        Where to redirect to after logging out.
    :param template_path:
        If you want to override the default logout HTML template, you can do
        so by specifying the absolute path to a custom template. For example
        ``'/some_directory/logout.html'``. Refer to the default template at
        ``piccolo_api/session_auth/templates/logout.html`` as a basis for your
        custom template.
    """
    template_path = (
        LOGOUT_TEMPLATE_PATH if template_path is None else template_path
    )

    directory, filename = os.path.split(template_path)
    environment = Environment(loader=FileSystemLoader(directory))
    logout_template = environment.get_template(filename)

    class _SessionLogoutEndpoint(SessionLogoutEndpoint):
        _session_table = session_table
        _cookie_name = cookie_name
        _redirect_to = redirect_to
        _logout_template = logout_template

    return _SessionLogoutEndpoint



from __future__ import annotations

import typing as t
from datetime import timedelta

from piccolo.apps.user.tables import BaseUser as PiccoloBaseUser
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    AuthenticationError,
    BaseUser,
)
from starlette.requests import HTTPConnection

from piccolo_api.session_auth.tables import SessionsBase
from piccolo_api.shared.auth import UnauthenticatedUser, User


class SessionsAuthBackend(AuthenticationBackend):
    """
    Authenticaion middleware which uses session cookies.
    """

    def __init__(
        self,
        auth_table: t.Type[PiccoloBaseUser] = PiccoloBaseUser,
        session_table: t.Type[SessionsBase] = SessionsBase,
        cookie_name: str = "id",
        admin_only: bool = True,
        superuser_only: bool = False,
        active_only: bool = True,
        increase_expiry: t.Optional[timedelta] = None,
        allow_unauthenticated: bool = False,
    ):
        """
        :param auth_table:
            The Piccolo table used for authenticating users.
        :param session_table:
            The Piccolo table used for storing sessions.
        :param cookie_name:
            The name of the session cookie. Override this if it clashes with
            other cookies in your application.
        :param admin_only:
            If True, users which aren't admins will be rejected.
        :param superuser_only:
            If True, users which aren't superusers will be rejected.
        :param active_only:
            If True, users which aren't active will be rejected.
        :param increase_expiry:
            If set, the session expiry will be increased by this amount on each
            request, if it's close to expiry. This allows sessions to have a
            short expiry date, whilst also providing a good user experience.
        :param allow_unauthenticated:
            If True, when a matching user session can't be found, the request
            still continues, but an unauthenticated user is added to the scope.
            It's then up to the application's endpoints to check if a user is
            authenticated or not using ``request.user.is_authenticated``. If
            False, the request is automatically rejected if a user session
            can't be found.
        """
        super().__init__()
        self.auth_table = auth_table
        self.session_table = session_table
        self.cookie_name = cookie_name
        self.admin_only = admin_only
        self.superuser_only = superuser_only
        self.active_only = active_only
        self.increase_expiry = increase_expiry
        self.allow_unauthenticated = allow_unauthenticated

    async def authenticate(
        self, conn: HTTPConnection
    ) -> t.Optional[t.Tuple[AuthCredentials, BaseUser]]:
        token = conn.cookies.get(self.cookie_name, None)
        if not token:
            if self.allow_unauthenticated:
                return (AuthCredentials(scopes=[]), UnauthenticatedUser())
            else:
                raise AuthenticationError("No session cookie found.")

        user_id = await self.session_table.get_user_id(
            token, increase_expiry=self.increase_expiry
        )

        if not user_id:
            if self.allow_unauthenticated:
                return (AuthCredentials(scopes=[]), UnauthenticatedUser())
            else:
                raise AuthenticationError("No matching session found.")

        piccolo_user = (
            await self.auth_table.objects()
            .where(self.auth_table._meta.primary_key == user_id)
            .first()
            .run()
        )

        if not piccolo_user:
            if self.allow_unauthenticated:
                return (AuthCredentials(scopes=[]), UnauthenticatedUser())
            else:
                raise AuthenticationError("That user doesn't exist anymore")

        if self.admin_only and not piccolo_user.admin:
            raise AuthenticationError("Admin users only")

        if self.superuser_only and not piccolo_user.superuser:
            raise AuthenticationError("Superusers only")

        if self.active_only and not piccolo_user.active:
            raise AuthenticationError("Active users only")

        user = User(user=piccolo_user)

        return (AuthCredentials(scopes=["authenticated"]), user)



from __future__ import annotations

import typing as t

from piccolo.apps.user.tables import BaseUser as PiccoloBaseUser
from starlette.authentication import BaseUser

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.table import Table


class User(BaseUser):
    def __init__(self, user: PiccoloBaseUser):
        super().__init__()
        self.user = user

    ###########################################################################
    # For backwards compatibility - these used to be arguments to the
    # contructor, but we can just infer them from the user instance.

    @property
    def auth_table(self) -> t.Type[Table]:
        return self.user.__class__

    @property
    def user_id(self) -> int:
        return t.cast(
            int, getattr(self.user, self.user._meta.primary_key._meta.name)
        )

    @property
    def username(self) -> str:
        return t.cast(str, self.user.username)

    ###########################################################################
    # Required properties.

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.username

    @property
    def identity(self) -> str:
        return str(self.user_id)


class UnauthenticatedUser(BaseUser):
    def __init__(self):
        super().__init__()
        self.user = None

    @property
    def is_authenticated(self) -> bool:
        return False

    @property
    def display_name(self) -> str:
        return ""

    @property
    def identity(self) -> str:
        return ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import find_packages, setup

from piccolo_api import __VERSION__ as VERSION

directory = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(directory, "requirements/requirements.txt")) as f:
    contents = f.read()
    REQUIREMENTS = [i.strip() for i in contents.strip().split("\n")]


with open(os.path.join(directory, "README.md")) as f:
    LONG_DESCRIPTION = f.read()


setup(
    name="piccolo_api",
    version=VERSION,
    description=(
        "Utilities for using the Piccolo ORM in ASGI apps, plus essential "
        "ASGI middleware such as authentication and rate limiting."
    ),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Daniel Townsend",
    author_email="dan@dantownsend.co.uk",
    python_requires=">=3.7.0",
    url="https://github.com/piccolo-orm/piccolo_api",
    packages=find_packages(exclude=("tests",)),
    package_data={
        "": [
            "templates/*",
        ],
        "piccolo_api": ["py.typed"],
    },
    install_requires=REQUIREMENTS,
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
)



import typing as t

import jwt
from piccolo.apps.user.tables import BaseUser
from starlette.exceptions import HTTPException


class JWTBlacklist:
    async def in_blacklist(self, token: str) -> bool:
        """
        Checks whether the token is in the blacklist.
        """
        return False


class JWTMiddleware:
    """
    Protects an endpoint - only allows access if a JWT token is presented.
    """

    auth_table: t.Type[BaseUser]

    def __init__(
        self,
        asgi,
        secret: str,
        auth_table: t.Type[BaseUser] = BaseUser,
        blacklist: JWTBlacklist = JWTBlacklist(),
    ) -> None:
        self.asgi = asgi
        self.secret = secret
        self.auth_table = auth_table
        self.blacklist = blacklist

    def get_token(self, headers: dict) -> t.Optional[str]:
        """
        Try and extract the JWT token from the request headers.
        """
        auth_token = headers.get(b"authorization", None)
        if not auth_token:
            return None
        auth_str = auth_token.decode()
        if not auth_str.startswith("Bearer "):
            return None
        return auth_str.split(" ")[1]

    async def get_user_id(
        self, token_dict: t.Dict[str, t.Any]
    ) -> t.Optional[int]:
        """
        Extract the user_id from the token, and check it's valid.
        """
        user_id = token_dict.get("user_id", None)

        if not user_id:
            return None

        exists = (
            await self.auth_table.exists()
            .where(self.auth_table._meta.primary_key == user_id)
            .run()
        )

        if exists is True:
            return user_id
        else:
            return None

    async def __call__(self, scope, receive, send):
        """
        Add the user_id to the scope if a JWT token is available, and the user
        is recognised, otherwise raise a 403 HTTP error.
        """
        headers = dict(scope["headers"])
        token = self.get_token(headers)
        if not token:
            raise HTTPException(status_code=403, detail="Token not found")

        if await self.blacklist.in_blacklist(token):
            raise HTTPException(status_code=403, detail="Token revoked")

        try:
            token_dict = jwt.decode(token, self.secret, algorithms=["HS256"])
        except jwt.exceptions.ExpiredSignatureError:
            raise HTTPException(status_code=403, detail="Token has expired")

        user_id = await self.get_user_id(token_dict)
        if not user_id:
            raise HTTPException(status_code=403)

        new_scope = dict(scope)
        new_scope["user_id"] = user_id

        await self.asgi(new_scope, receive, send)



from __future__ import annotations

import secrets
import typing as t
from datetime import datetime, timedelta

from piccolo.columns import Integer, Timestamp, Varchar
from piccolo.columns.defaults.timestamp import TimestampOffset
from piccolo.table import Table
from piccolo.utils.sync import run_sync


class SessionsBase(Table, tablename="sessions"):
    """
    Use this table, or inherit from it, to create for a session store.

    We set a hard limit on the expiry date - it can keep on getting extended
    up until this value, after which it's best to invalidate it, and either
    require login again, or just create a new session token.
    """

    token = Varchar(length=100, null=False)
    user_id = Integer(null=False)
    expiry_date: t.Union[Timestamp, datetime] = Timestamp(
        default=TimestampOffset(hours=1), null=False
    )
    max_expiry_date: t.Union[Timestamp, datetime] = Timestamp(
        default=TimestampOffset(days=7), null=False
    )

    @classmethod
    async def create_session(
        cls,
        user_id: int,
        expiry_date: t.Optional[datetime] = None,
        max_expiry_date: t.Optional[datetime] = None,
    ) -> SessionsBase:
        while True:
            token = secrets.token_urlsafe(nbytes=32)
            if not await cls.exists().where(cls.token == token).run():
                break

        session = cls(token=token, user_id=user_id)
        if expiry_date:
            session.expiry_date = expiry_date
        if max_expiry_date:
            session.max_expiry_date = max_expiry_date

        await session.save().run()

        return session

    @classmethod
    def create_session_sync(
        cls, user_id: int, expiry_date: t.Optional[datetime] = None
    ) -> SessionsBase:
        return run_sync(cls.create_session(user_id, expiry_date))

    @classmethod
    async def get_user_id(
        cls, token: str, increase_expiry: t.Optional[timedelta] = None
    ) -> t.Optional[int]:
        """
        Returns the user_id if the given token is valid, otherwise None.

        :param increase_expiry:
            If set, the `expiry_date` will be increased by the given amount
            if it's close to expiring. If it has already expired, nothing
            happens. The `max_expiry_date` remains the same, so there's a hard
            limit on how long a session can be used for.
        """
        session: SessionsBase = (
            await cls.objects().where(cls.token == token).first().run()
        )

        if not session:
            return None

        now = datetime.now()
        if (session.expiry_date > now) and (session.max_expiry_date > now):
            if increase_expiry and (
                t.cast(datetime, session.expiry_date) - now < increase_expiry
            ):
                session.expiry_date = (
                    t.cast(datetime, session.expiry_date) + increase_expiry
                )
                await session.save().run()

            return t.cast(t.Optional[int], session.user_id)
        else:
            return None

    @classmethod
    def get_user_id_sync(cls, token: str) -> t.Optional[int]:
        return run_sync(cls.get_user_id(token))

    @classmethod
    async def remove_session(cls, token: str):
        await cls.delete().where(cls.token == token).run()

    @classmethod
    def remove_session_sync(cls, token: str):
        return run_sync(cls.remove_session(token))



import os

from piccolo.conf.apps import AppConfig

from .commands import clean
from .tables import SessionsBase

CURRENT_DIRECTORY = os.path.dirname(os.path.abspath(__file__))


APP_CONFIG = AppConfig(
    app_name="session_auth",
    migrations_folder_path=os.path.join(
        CURRENT_DIRECTORY, "piccolo_migrations"
    ),
    table_classes=[SessionsBase],
    migration_dependencies=[],
    commands=[clean],
)
